from queue import Queue
from threading import Thread

from naneos.logger.custom_logger import get_naneos_logger
from naneos.partector import Partector2
from naneos.serial_utils import list_serial_ports as ls_ports

logger = get_naneos_logger(__name__)


def scan_for_serial_partector(serial_number: int) -> str:
    """Scans all possible ports using threads (fast)."""
    threads = []
    q = Queue()

    [threads.append(Thread(target=__scan_port, args=(port, q))) for port in ls_ports()]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    ports = {k: v for d in tuple(q.queue) for (k, v) in d.items()}
    logger.debug(f"Found ports: {ports}")

    if serial_number in ports.keys():
        return ports[serial_number]

    return None


def scan_for_serial_partectors(sn_exclude: list = []) -> dict:
    """Scans all possible ports using threads (fast)."""
    threads = []
    q = Queue()

    [
        threads.append(Thread(target=__scan_port, args=(port, q)))
        for port in ls_ports()
        if port not in sn_exclude
    ]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    p1 = {k: v for x in tuple(q.queue) for (k, v) in x.items() if k < 1000}
    p2 = {k: v for x in tuple(q.queue) for (k, v) in x.items() if k >= 1000}

    return {"P1": p1, "P2": p2}


def __scan_port(port: str, q: Queue) -> dict:
    try:
        p2 = Partector2(port=port)
        q.put({p2._serial_number: port})
        p2.close(blocking=True)
    except Exception:
        try:
            p2.close(blocking=True)
        except UnboundLocalError:
            pass  # p2 object already destroyed
        except Exception:
            raise Exception(f"Could not close port {port}.")



import base64
import datetime
from typing import Optional

import pandas as pd

import naneos.protobuf.protoV1_pb2 as pbScheme


def create_combined_entry(
    devices: list[pbScheme.Device],
    abs_timestamp: int,
    gateway_points: Optional[list[pbScheme.GatewayPoint]] = None,
    position_points: Optional[list[pbScheme.PositionPoint]] = None,
    wind_points: Optional[list[pbScheme.WindPoint]] = None,
) -> pbScheme.CombinedData:
    combined = pbScheme.CombinedData()
    combined.abs_timestamp = abs_timestamp

    combined.devices.extend(devices)

    if gateway_points is not None:
        combined.gateway_points.extend(gateway_points)

    if position_points is not None:
        combined.position_points.extend(position_points)

    if wind_points is not None:
        combined.wind_points.extend(wind_points)

    return combined


def create_partector1_entry(
    df: pd.DataFrame, serial_number: int, abs_time: int
) -> pbScheme.Device:
    device = pbScheme.Device()

    device.type = 0  # Partector1
    device.serial_number = serial_number

    for _, row in df.iterrows():
        device_point = pbScheme.DevicePoint()

        device_point.timestamp = abs_time - int(row.name.timestamp())
        device_point.device_status = int(row["device_status"])

        device_point.ldsa = int(row["LDSA"] * 100.0)
        device_point.battery_voltage = int(row["batt_voltage"] * 100.0)

        idiff_tmp = row["idiff_global"] if row["idiff_global"] > 0 else 0
        device_point.diffusion_current = int(idiff_tmp * 100.0)
        device_point.corona_voltage = int(row["ucor_global"])
        device_point.flow = int(row["flow"] * 1000.0)
        device_point.temperature = int(row["T"])
        device_point.relative_humidity = int(row["RHcorr"])

        device.device_points.append(device_point)

    return device


def create_partector_2_pro_garagenbox(
    df: pd.DataFrame, serial_number: int, abs_time: int
) -> pbScheme.Device:
    if len(df) == 0:
        raise ValueError("Dataframe must not be empty!")

    device = pbScheme.Device()  # contains type, serial_number, device_points
    device.type = 0  # Partector1/2 #TODO: ask @rschre what he wants for special devices
    device.serial_number = serial_number

    # TODO: check if using the right axis
    device_points = df.apply(_create_device_point, axis=1, abs_time=abs_time).to_list()
    device_points = [x for x in device_points if x is not None]

    device.device_points.extend(device_points)

    return device


def _create_device_point(ser: pd.Series, abs_time: int) -> Optional[pbScheme.DevicePoint]:
    try:
        device_point = pbScheme.DevicePoint()

        # mandatory fields
        device_point.timestamp = abs_time - ser.name
        device_point.device_status = int(ser["device_status"])

        # optional fields
        device_point.particle_number_concentration = int(ser["number"])
        device_point.average_particle_diameter = int(ser["diameter"])
        device_point.ldsa = int(ser["LDSA"] * 100.0)
        device_point.surface = int(ser["surface"] * 100.0)
        device_point.particle_mass = int(ser["particle_mass"] * 100.0)
        device_point.sigma_size_dist = int(ser["sigma"] * 100.0)
        idiff_tmp = ser["idiff_global"] if ser["idiff_global"] > 0 else 0
        device_point.diffusion_current = int(idiff_tmp * 100.0)
        device_point.corona_voltage = int(ser["ucor_global"])
        device_point.deposition_voltage = int(ser["deposition_voltage"])
        device_point.temperature = int(ser["T"])
        device_point.relative_humidity = int(ser["RHcorr"])
        device_point.ambient_pressure = int(ser["P_average"] * 10.0)
        device_point.flow = int(ser["flow_from_dp"] * 1000.0)
        device_point.battery_voltage = int(ser["batt_voltage"] * 100.0)
        device_point.pump_current = int(ser["pump_current"] * 100.0)
        device_point.pump_pwm = int(ser["pump_pwm"])

        # pro stuff
        device_point.cs_status = int(ser["cs_status"])
        device_point.steps_inversion = int(ser["steps"])
        device_point.current_dist_0 = int(ser["current_0"])
        device_point.current_dist_1 = int(ser["current_1"])
        device_point.current_dist_2 = int(ser["current_2"])
        device_point.current_dist_3 = int(ser["current_3"])
        device_point.current_dist_4 = int(ser["current_4"])
        device_point.particle_number_10nm = int(ser["particle_number_10nm"])
        device_point.particle_number_16nm = int(ser["particle_number_16nm"])
        device_point.particle_number_26nm = int(ser["particle_number_26nm"])
        device_point.particle_number_43nm = int(ser["particle_number_43nm"])
        device_point.particle_number_70nm = int(ser["particle_number_70nm"])
        device_point.particle_number_114nm = int(ser["particle_number_114nm"])
        device_point.particle_number_185nm = int(ser["particle_number_185nm"])
        device_point.particle_number_300nm = int(ser["particle_number_300nm"])
    except Exception as e:
        print(f"Error in _create_device_Point: {e}")
        return None

    return device_point


if __name__ == "__main__":
    df = pd.read_pickle("/Users/huegi/gitlocal/naneos/naneos-devices/tests/df_garagae.pkl")

    print(df)

    abs_time = int(datetime.datetime.now().timestamp())
    serial_number = 777
    device = create_partector_2_pro_garagenbox(df, serial_number, abs_time)

    print(device)

    # df = pd.read_pickle("p1.pkl")

    # abs_time = int(datetime.datetime.now().timestamp())

    # device_list = []

    # device_list.append(create_Partector1_entry(df, 16, abs_time))

    # combined = create_Combined_entry(devices=device_list, abs_time=abs_time)

    # json_string = pbJson.MessageToJson(combined, including_default_value_fields=True)
    # proto_string = combined.SerializeToString()

    # # print(json_string)
    # # print(proto_string)

    # # base64 encode proto_string
    # base64_string = base64.b64encode(proto_string)

    # print(base64_string)



import logging
from typing import Optional


class CustomFormatter(logging.Formatter):
    _grey = "\x1b[38;20m"
    _green = "\x1b[32;20m"
    _yellow = "\x1b[33;20m"
    _red = "\x1b[31;20m"
    _bold_red = "\x1b[31;1m"
    _reset = "\x1b[0m"
    _format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    _FORMATS_TERMINAL: tuple[tuple[int, str], ...] = (
        (logging.DEBUG, _grey + _format + _reset),
        (logging.INFO, _green + _format + _reset),
        (logging.WARNING, _yellow + _format + _reset),
        (logging.ERROR, _red + _format + _reset),
        (logging.CRITICAL, _bold_red + _format + _reset),
    )
    _FORMATS_SAVE: tuple[tuple[int, str], ...] = (
        (logging.DEBUG, _format),
        (logging.INFO, _format),
        (logging.WARNING, _format),
        (logging.ERROR, _format),
        (logging.CRITICAL, _format),
    )

    def __init__(self, terminal: bool = False, fmt: Optional[str] = None):
        self._FORMATS = self._FORMATS_TERMINAL if terminal else self._FORMATS_SAVE
        if fmt is None:
            fmt = self._format
        super().__init__(fmt=fmt)

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = next(
            (x[1] for x in self._FORMATS if x[0] == record.levelno),
            self._FORMATS[0][1],
        )

        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_naneos_logger(name: str, level: int = logging.INFO) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter_file = CustomFormatter(terminal=False)
    file_handler = logging.FileHandler("naneos-devices.log")
    file_handler.setLevel(level)
    file_handler.setFormatter(formatter_file)

    formatter_stream = CustomFormatter(terminal=True)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(formatter_stream)

    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)

    return logger


if __name__ == "__main__":
    logger = get_naneos_logger(__name__, logging.DEBUG)
    logger.debug("debug message")
    logger.info("info message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")



from pathlib import Path
import sys

import serial


def list_serial_ports() -> list[str]:
    """Returns a list of serial ports available on the system.

    Raises:
        OSError: If the platform is not supported.

    Returns:
        list[str]: A list of serial ports available on the system.
    """
    ports: list[str] = _get_all_open_ports()
    ports = _check_port_function(ports)

    return ports


def _get_all_open_ports() -> list[str]:
    ports: list[str] = []

    if sys.platform.startswith("win"):
        ports = ["COM%s" % (i + 1) for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        ports = [str(port) for port in Path("/dev").glob("tty*") if "tty." not in str(port)]
    elif sys.platform.startswith("darwin"):  # mac
        ports = [str(port) for port in Path("/dev").glob("tty.*") if "Bluetooth" not in str(port)]
    else:
        raise OSError(f"Unsupported platform: {sys.platform}")

    return ports


def _check_port_function(ports: list[str]) -> list[str]:
    working_ports: list[str] = []

    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            working_ports.append(port)
        except (OSError, serial.SerialException):
            pass

    return working_ports



from threading import Event, Thread

from naneos.logger.custom_logger import get_naneos_logger

logger = get_naneos_logger(__name__)


class PartectorCheckerThread(Thread):
    # from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint

    def __init__(self, partector) -> None:
        super().__init__()
        self.partector = partector

        self._stop_event = Event()
        self.start()

    def stop(self) -> None:
        self._stop_event.set()

    def run(self) -> None:
        # every 2.1 seconds check if device is still connected
        while not self._stop_event.wait(2.1):
            try:
                self.partector._run_check_connection()
            except Exception as e:
                logger.error(e)



from naneos.partector.blueprints._data_structure import PARTECTOR1_DATA_STRUCTURE
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint


class Partector1(PartectorBluePrint):
    def __init__(self, port: str, verb_freq: int = 1) -> None:
        super().__init__(port, verb_freq)

    def _init_serial_data_structure(self):
        self._data_structure = PARTECTOR1_DATA_STRUCTURE

    def _set_verbose_freq(self, freq: int):
        """
        Set the frequency of the verbose output.

        :param int freq: Frequency of the verbose output in Hz. (0: off, 1: 1Hz, 2: 10Hz, 3: 100Hz)
        """

        if freq < 0 or freq > 3:
            raise ValueError("Frequency must be between 0 and 3!")

        self._write_line(f"X000{freq}!")



from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers

DESCRIPTOR: _descriptor.FileDescriptor

class CombinedData(_message.Message):
    __slots__ = ("abs_timestamp", "devices", "gateway_points", "position_points", "wind_points")
    ABS_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DEVICES_FIELD_NUMBER: _ClassVar[int]
    GATEWAY_POINTS_FIELD_NUMBER: _ClassVar[int]
    POSITION_POINTS_FIELD_NUMBER: _ClassVar[int]
    WIND_POINTS_FIELD_NUMBER: _ClassVar[int]
    abs_timestamp: int
    devices: _containers.RepeatedCompositeFieldContainer[Device]
    gateway_points: _containers.RepeatedCompositeFieldContainer[GatewayPoint]
    position_points: _containers.RepeatedCompositeFieldContainer[PositionPoint]
    wind_points: _containers.RepeatedCompositeFieldContainer[WindPoint]
    def __init__(
        self,
        abs_timestamp: _Optional[int] = ...,
        devices: _Optional[_Iterable[_Union[Device, _Mapping]]] = ...,
        gateway_points: _Optional[_Iterable[_Union[GatewayPoint, _Mapping]]] = ...,
        position_points: _Optional[_Iterable[_Union[PositionPoint, _Mapping]]] = ...,
        wind_points: _Optional[_Iterable[_Union[WindPoint, _Mapping]]] = ...,
    ) -> None: ...

class Device(_message.Message):
    __slots__ = ("type", "serial_number", "device_points")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    DEVICE_POINTS_FIELD_NUMBER: _ClassVar[int]
    type: int
    serial_number: int
    device_points: _containers.RepeatedCompositeFieldContainer[DevicePoint]
    def __init__(
        self,
        type: _Optional[int] = ...,
        serial_number: _Optional[int] = ...,
        device_points: _Optional[_Iterable[_Union[DevicePoint, _Mapping]]] = ...,
    ) -> None: ...

class DevicePoint(_message.Message):
    __slots__ = (
        "timestamp",
        "device_status",
        "ldsa",
        "average_particle_diameter",
        "particle_number_concentration",
        "temperature",
        "relative_humidity",
        "battery_voltage",
        "particle_mass",
        "corona_voltage",
        "diffusion_current",
        "deposition_voltage",
        "flow",
        "ambient_pressure",
        "electrometer_offset",
        "electrometer_2_offset",
        "electrometer_gain",
        "electrometer_2_gain",
        "diffusion_current_offset",
        "particle_number_10nm",
        "particle_number_16nm",
        "particle_number_26nm",
        "particle_number_43nm",
        "particle_number_70nm",
        "particle_number_114nm",
        "particle_number_185nm",
        "particle_number_300nm",
        "surface",
        "sigma_size_dist",
        "steps_inversion",
        "current_dist_0",
        "current_dist_1",
        "current_dist_2",
        "current_dist_3",
        "current_dist_4",
        "pump_current",
        "pump_pwm",
        "cs_status",
    )
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DEVICE_STATUS_FIELD_NUMBER: _ClassVar[int]
    LDSA_FIELD_NUMBER: _ClassVar[int]
    AVERAGE_PARTICLE_DIAMETER_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_CONCENTRATION_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    RELATIVE_HUMIDITY_FIELD_NUMBER: _ClassVar[int]
    BATTERY_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_MASS_FIELD_NUMBER: _ClassVar[int]
    CORONA_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_FIELD_NUMBER: _ClassVar[int]
    DEPOSITION_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    FLOW_FIELD_NUMBER: _ClassVar[int]
    AMBIENT_PRESSURE_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_OFFSET_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_2_OFFSET_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_GAIN_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_2_GAIN_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_OFFSET_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_10NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_16NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_26NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_43NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_70NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_114NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_185NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_300NM_FIELD_NUMBER: _ClassVar[int]
    SURFACE_FIELD_NUMBER: _ClassVar[int]
    SIGMA_SIZE_DIST_FIELD_NUMBER: _ClassVar[int]
    STEPS_INVERSION_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_0_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_1_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_2_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_3_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_4_FIELD_NUMBER: _ClassVar[int]
    PUMP_CURRENT_FIELD_NUMBER: _ClassVar[int]
    PUMP_PWM_FIELD_NUMBER: _ClassVar[int]
    CS_STATUS_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    device_status: int
    ldsa: int
    average_particle_diameter: int
    particle_number_concentration: int
    temperature: int
    relative_humidity: int
    battery_voltage: int
    particle_mass: int
    corona_voltage: int
    diffusion_current: int
    deposition_voltage: int
    flow: int
    ambient_pressure: int
    electrometer_offset: int
    electrometer_2_offset: int
    electrometer_gain: int
    electrometer_2_gain: int
    diffusion_current_offset: int
    particle_number_10nm: int
    particle_number_16nm: int
    particle_number_26nm: int
    particle_number_43nm: int
    particle_number_70nm: int
    particle_number_114nm: int
    particle_number_185nm: int
    particle_number_300nm: int
    surface: int
    sigma_size_dist: int
    steps_inversion: int
    current_dist_0: int
    current_dist_1: int
    current_dist_2: int
    current_dist_3: int
    current_dist_4: int
    pump_current: int
    pump_pwm: int
    cs_status: int
    def __init__(
        self,
        timestamp: _Optional[int] = ...,
        device_status: _Optional[int] = ...,
        ldsa: _Optional[int] = ...,
        average_particle_diameter: _Optional[int] = ...,
        particle_number_concentration: _Optional[int] = ...,
        temperature: _Optional[int] = ...,
        relative_humidity: _Optional[int] = ...,
        battery_voltage: _Optional[int] = ...,
        particle_mass: _Optional[int] = ...,
        corona_voltage: _Optional[int] = ...,
        diffusion_current: _Optional[int] = ...,
        deposition_voltage: _Optional[int] = ...,
        flow: _Optional[int] = ...,
        ambient_pressure: _Optional[int] = ...,
        electrometer_offset: _Optional[int] = ...,
        electrometer_2_offset: _Optional[int] = ...,
        electrometer_gain: _Optional[int] = ...,
        electrometer_2_gain: _Optional[int] = ...,
        diffusion_current_offset: _Optional[int] = ...,
        particle_number_10nm: _Optional[int] = ...,
        particle_number_16nm: _Optional[int] = ...,
        particle_number_26nm: _Optional[int] = ...,
        particle_number_43nm: _Optional[int] = ...,
        particle_number_70nm: _Optional[int] = ...,
        particle_number_114nm: _Optional[int] = ...,
        particle_number_185nm: _Optional[int] = ...,
        particle_number_300nm: _Optional[int] = ...,
        surface: _Optional[int] = ...,
        sigma_size_dist: _Optional[int] = ...,
        steps_inversion: _Optional[int] = ...,
        current_dist_0: _Optional[int] = ...,
        current_dist_1: _Optional[int] = ...,
        current_dist_2: _Optional[int] = ...,
        current_dist_3: _Optional[int] = ...,
        current_dist_4: _Optional[int] = ...,
        pump_current: _Optional[int] = ...,
        pump_pwm: _Optional[int] = ...,
        cs_status: _Optional[int] = ...,
    ) -> None: ...

class GatewayPoint(_message.Message):
    __slots__ = (
        "timestamp",
        "serial_number",
        "firmware_version",
        "free_memory",
        "free_heap",
        "largest_free_block_heap",
        "cellular_signal",
        "battery_int_soc",
        "battery_ext_soc",
        "battery_ext_voltage",
        "charging_ext_voltage",
    )
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    FIRMWARE_VERSION_FIELD_NUMBER: _ClassVar[int]
    FREE_MEMORY_FIELD_NUMBER: _ClassVar[int]
    FREE_HEAP_FIELD_NUMBER: _ClassVar[int]
    LARGEST_FREE_BLOCK_HEAP_FIELD_NUMBER: _ClassVar[int]
    CELLULAR_SIGNAL_FIELD_NUMBER: _ClassVar[int]
    BATTERY_INT_SOC_FIELD_NUMBER: _ClassVar[int]
    BATTERY_EXT_SOC_FIELD_NUMBER: _ClassVar[int]
    BATTERY_EXT_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    CHARGING_EXT_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    serial_number: int
    firmware_version: int
    free_memory: int
    free_heap: int
    largest_free_block_heap: int
    cellular_signal: int
    battery_int_soc: int
    battery_ext_soc: int
    battery_ext_voltage: int
    charging_ext_voltage: int
    def __init__(
        self,
        timestamp: _Optional[int] = ...,
        serial_number: _Optional[int] = ...,
        firmware_version: _Optional[int] = ...,
        free_memory: _Optional[int] = ...,
        free_heap: _Optional[int] = ...,
        largest_free_block_heap: _Optional[int] = ...,
        cellular_signal: _Optional[int] = ...,
        battery_int_soc: _Optional[int] = ...,
        battery_ext_soc: _Optional[int] = ...,
        battery_ext_voltage: _Optional[int] = ...,
        charging_ext_voltage: _Optional[int] = ...,
    ) -> None: ...

class PositionPoint(_message.Message):
    __slots__ = ("timestamp", "latitude", "longitude")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    LATITUDE_FIELD_NUMBER: _ClassVar[int]
    LONGITUDE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    latitude: float
    longitude: float
    def __init__(
        self,
        timestamp: _Optional[int] = ...,
        latitude: _Optional[float] = ...,
        longitude: _Optional[float] = ...,
    ) -> None: ...

class WindPoint(_message.Message):
    __slots__ = ("timestamp", "wind_speed", "wind_angle")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    WIND_SPEED_FIELD_NUMBER: _ClassVar[int]
    WIND_ANGLE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    wind_speed: int
    wind_angle: int
    def __init__(
        self,
        timestamp: _Optional[int] = ...,
        wind_speed: _Optional[int] = ...,
        wind_angle: _Optional[int] = ...,
    ) -> None: ...



from logging import CRITICAL as LEVEL_CRITICAL
from logging import DEBUG as LEVEL_DEBUG
from logging import ERROR as LEVEL_ERROR
from logging import INFO as LEVEL_INFO
from logging import WARNING as LEVEL_WARNING

from naneos.logger.custom_logger import get_naneos_logger

__all__ = [
    "get_naneos_logger",
    "LEVEL_DEBUG",
    "LEVEL_INFO",
    "LEVEL_WARNING",
    "LEVEL_ERROR",
    "LEVEL_CRITICAL",
]



from .protobuf import (
    create_combined_entry,
    create_partector1_entry,
    create_partector_2_pro_garagenbox,
)

__all__ = [
    "create_partector1_entry",
    "create_combined_entry",
    "create_partector_2_pro_garagenbox",
]



from datetime import datetime, timezone
import stat
from typing import Any, Optional

from naneos.logger.custom_logger import get_naneos_logger
from naneos.partector.blueprints._data_structure import PARTECTOR2_PRO_GARAGE_DATA_STRUCTURE
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint

logger = get_naneos_logger(__name__)


class Partector2ProGarage(PartectorBluePrint):
    CS_OFF = 0
    CS_ON = 1
    CS_UNKNOWN = -1

    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
        **kwargs: Any,
    ) -> None:
        self._catalyst_state = self.CS_UNKNOWN
        self._auto_mode = True

        self._callback_catalyst = kwargs.get("callback_catalyst", None)
        if self._callback_catalyst is None:
            logger.error("No callback function for catalyst state given!")
            raise ValueError("No callback function for catalyst state given!")
        super().__init__(serial_number, port, verb_freq)

    def _init_serial_data_structure(self) -> None:
        self._data_structure = PARTECTOR2_PRO_GARAGE_DATA_STRUCTURE

    def _set_verbose_freq(self, freq: int) -> None:
        if freq == 0:
            self._write_line("X0000!")
        else:
            self._write_line("X0006!")

    def set_catalyst_state(self, state: str) -> None:
        """Sets the catalyst state to on, off or auto."""
        if not self._connected:
            return

        if state == "on":
            self._write_line("CSon!")
            self._cs_state = self.CS_ON
            self._auto_mode = False
        elif state == "off":
            self._write_line("CSoff!")
            self._cs_state = self.CS_OFF
            self._auto_mode = False
        elif state == "auto":
            self._write_line("CSauto!")
            self._auto_mode = True
        else:
            logger.warning(f"Unknown catalyst state: {state} -> nothing done.")
            return

        logger.info(f"Catalyst state set to {state}.")

    def _serial_reading_routine(self) -> None:
        line = self._read_line()

        if not line or line == "":
            return

        # check if line contains !CS_on or !CS_off and remove it from line
        if "!CS_on" in line:
            line = line.replace("!CS_on", "")
            self._callback_catalyst(True)
            self._put_line_to_queue(line, True)
            self._catalyst_state = self.CS_ON
        elif "!CS_off" in line:
            line = line.replace("!CS_off", "")
            self._callback_catalyst(False)
            self._put_line_to_queue(line, False)
            self._catalyst_state = self.CS_OFF
        else:
            self._put_line_to_queue(line)

    def _put_line_to_queue(self, line: str, cs_command: Optional[bool] = None) -> None:
        unix_timestamp = int(datetime.now(tz=timezone.utc).timestamp())
        data = [unix_timestamp] + line.split("\t")

        if len(data) != len(self._data_structure):
            self._put_to_info_queue(data)
            return

        state: Optional[int] = None
        try:
            state = int(data[-1])
        except Exception as excep:
            logger.warning(f"Could not parse catalyst state in backup function: {excep}")

        if state in [0, 1] and self._catalyst_state != state:
            self._callback_catalyst(state == 1)
            self._catalyst_state = state
            logger.warning(f"Set catalyst state to by backup function to {state}.")

        # removes the size dist from all the measurements that are not wanted
        if cs_command is None and self._auto_mode:
            data[20:28] = [0] * 8

        self._put_to_data_queue(data)

    def _put_to_info_queue(self, data: list[int | str]) -> None:
        if self._queue_info.full():
            self._queue_info.get()
        self._queue_info.put(data)

    def _put_to_data_queue(self, data: list[int | str]) -> None:
        if self._queue.full():
            self._queue.get()
        self._queue.put(data)


if __name__ == "__main__":
    import time

    def test_callback(state: bool) -> None:
        logger.info(f"Catalyst state changed to {state}.")

    p2 = Partector2ProGarage(serial_number=8440, callback_catalyst=test_callback)

    # print(p2.write_line("v?", 1))
    time.sleep(5)

    df = p2.get_data_pandas()
    print(df)
    df.to_pickle("tests/df_garagae.pkl")

    print("Closing...")
    p2.close(blocking=True)
    print("Closed!")



from .partector1 import Partector1
from .partector2 import Partector2
from .partector2_pro import Partector2Pro
from .partector2_pro_garage import Partector2ProGarage
from .scan_for_partector import scan_for_serial_partector, scan_for_serial_partectors

__all__ = [
    "Partector1",
    "Partector2",
    "Partector2Pro",
    "Partector2ProGarage",
    "scan_for_serial_partector",
    "scan_for_serial_partectors",
]



from naneos.partector.blueprints._data_structure import PARTECTOR2_PRO_DATA_STRUCTURE
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint


class Partector2Pro(PartectorBluePrint):
    def __init__(self, port: str, verb_freq: int = 1) -> None:
        super().__init__(port, verb_freq)

    def _init_serial_data_structure(self):
        self._data_structure = PARTECTOR2_PRO_DATA_STRUCTURE

    def _set_verbose_freq(self, freq: int):
        if freq == 0:
            self._write_line("X0000!")
        else:
            self._write_line("X0004!")


if __name__ == "__main__":
    import time

    p2 = Partector2Pro("/dev/tty.usbmodemDOSEMet_1")

    print(p2.write_line("v?", 1))
    time.sleep(2)
    print(p2.get_data_pandas()["T"])
    p2.close()



import asyncio
import base64
import datetime
from queue import Queue
from threading import Event, Thread
import time

import requests

from naneos.partector import Partector1, scan_for_serial_partectors
from naneos.protobuf import create_combined_entry, create_partector1_entry


class P1uploadThread(Thread):
    def __init__(self) -> None:
        Thread.__init__(self)
        self.event = Event()
        self.event.set()
        self.connect_queue = Queue()
        self.disconnect_queue = Queue()

        self.device_list = []
        self.last_send_time = time.time()  # change to asyncio loop
        self.start()

    def run(self) -> None:
        asyncio.run(self.run_async())

        print("closing devices")
        for d in self.device_list:
            d.close()

    async def run_async(self) -> None:
        task1 = asyncio.create_task(self.connect_disconnect_routine())
        task2 = asyncio.create_task(self._send_data_to_server())

        await task1
        await task2

    async def connect_disconnect_routine(self):
        while self.event.is_set():
            self._connect_to_ports_from_queue()
            self._disconnect_from_void_ports()
            self._disconnect_from_ports_from_queue()

            await asyncio.sleep(0.1)

    def get_connected_devices_ports(self):
        return [d._ser.port for d in self.device_list]

    def _connect_to_ports_from_queue(self):
        connected_ports = self.get_connected_devices_ports()

        while not self.connect_queue.empty():
            port = self.connect_queue.get()
            if port not in connected_ports:
                print(f"Connecting to {port}")
                self.device_list.append(Partector1(port))

    def _disconnect_from_void_ports(self):
        for d in self.device_list:
            # print(d.last_heard)
            if d.last_heard < time.time() - 10:
                print(f"Disconnecting from {d._ser.port}")
                d.close()
                self.device_list.remove(d)

    def _disconnect_from_ports_from_queue(self):
        connected_ports = self.get_connected_devices_ports()

        while not self.disconnect_queue.empty():
            port = self.disconnect_queue.get()
            if port in connected_ports:
                # close the device in list and remove it from the list
                for d in self.device_list:
                    if d._ser.port == port:
                        d.close()
                        self.device_list.remove(d)

    async def _send_data_to_server(self):
        while self.event.is_set():
            # print("Sending data to server")
            await asyncio.sleep(15)

            if len(self.device_list) == 0:
                continue

            abs_time = int(datetime.datetime.now().timestamp())
            devices = []
            for d in self.device_list:
                df = d.get_data_pandas()
                devices.append(create_partector1_entry(df, d._serial_number, abs_time))

            combined = create_combined_entry(devices=devices, abs_timestamp=abs_time)

            proto_string = combined.SerializeToString()
            proto_string_base64 = base64.b64encode(proto_string)
            # print(proto_string_base64)

            # send this string to the server
            url = "https://hg3zkburji.execute-api.eu-central-1.amazonaws.com/prod/proto/v1"
            headers = {"Content-Type": "application/json", "Accept": "application/json"}
            body = f"""
                    {{
                        "gateway": "python_webhook",
                        "data": "{proto_string_base64.decode()}",
                        "published_at": "{datetime.datetime.now().isoformat()}"
                    }}
                    """

            # print(body)

            # send the data to the server
            r = requests.post(url, headers=headers, data=body)
            print(f"Status code: {r.status_code}, text: {r.text}")


if __name__ == "__main__":
    upload_thread = P1uploadThread()
    time.sleep(1)

    # wait for keyboard interrupt
    try:
        while True:
            time.sleep(1)
            connected_ports = upload_thread.get_connected_devices_ports()
            devs = scan_for_serial_partectors(sn_exclude=connected_ports)["P1"]
            # print(devs)

            # add all devices to the queue
            for v in devs.values():
                print(f"Adding {v} to queue")
                upload_thread.connect_queue.put(v)
    except KeyboardInterrupt:
        pass

    # stop the thread
    upload_thread.event.clear()
    upload_thread.join()



import asyncio
from collections import deque
from threading import Event, Thread
import time
from typing import Optional

from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from naneos.logger import LEVEL_DEBUG, get_naneos_logger
from naneos.partector_ble.partector_ble_device import PartectorBleDevice

logger = get_naneos_logger(__name__, LEVEL_DEBUG)


class PartectorBle(Thread):
    SERVICE_UUID = "0bd51666-e7cb-469b-8e4d-2742f1ba77cc"
    CHAR_STD = "e7add780-b042-4876-aae1-112855353cc1"
    CHAR_AUX = "e7add781-b042-4876-aae1-112855353cc1"
    CHAR_WRITE = "e7add782-b042-4876-aae1-112855353cc1"
    CHAR_READ = "e7add783-b042-4876-aae1-112855353cc1"
    CHAR_SIZE_DIST = "e7add784-b042-4876-aae1-112855353cc1"

    def __init__(self, serial_numbers: Optional[list[int]] = None) -> None:
        super().__init__()
        self.event = Event()

        self._serial_numbers = serial_numbers

        """Dict of all devices that are found in the scan method."""
        self._devices_to_check: dict[BLEDevice, tuple[int, AdvertisementData]] = {}
        """Dict containing all connected devices, that are in use."""
        self._partector_clients: dict[int, PartectorBleDevice] = {}

        self._send_queue: deque[tuple[int, str]] = deque(maxlen=10)

    def run(self) -> None:
        """Main loop of the partector BLE thread"""
        # asyncio.run(self.async_run())

        # run async_run and async_send_command in the same thread
        async def test() -> None:
            await asyncio.gather(self._async_run(), self._async_send_command())

        asyncio.run(test())

    async def _async_run(self) -> None:
        """Async implementation of the main loop of the partector BLE thread

        This is needed to use BleakScanner and BleakClient in the same thread
        """
        timestamp = int(time.time())  # unix timestamp in seconds
        while not self.event.is_set():
            if time.time() >= timestamp + 1:
                timestamp = int(time.time())

            await self.scan()
            await self.check_scans()

        await self._disconnect()

    async def _async_send_command(self) -> None:
        """Runs parallel to the scan an get functions and sends if somethin is pushed into the send queue"""
        while not self.event.is_set():
            if self._send_queue:
                serial_number, command = self._send_queue.popleft()
                if serial_number not in self._partector_clients:
                    logger.warning(f"Serial number {serial_number} not found")
                    continue

                client = self._partector_clients[serial_number].ble_client
                if client is None:
                    logger.warning(f"Serial number {serial_number} not connected")
                    continue
                else:
                    await client.write_gatt_char(self.CHAR_WRITE, command.encode())
                    logger.info(f"Sent {command} to {serial_number}")

            await asyncio.sleep(0.1)

    def send_command(self, serial_number: int, command: str) -> None:
        self._send_queue.append((serial_number, command))

    async def scan(self) -> None:
        """Scans for old and new Partector BLE protocol devices

        All found devices are added to the _devices_to_check dict in the _detection_callback method.
        Scan duration of 850ms is used because the advertised data of the old devices changes every 1s.
        """
        async with BleakScanner(detection_callback=self._detection_callback) as scanner:
            await scanner.start()
            await asyncio.sleep(0.85)
            await scanner.stop()

    async def check_scans(self) -> None:
        """Tries to connect to all devices in the _devices_to_connect dict.

        If the device is a old one, the advertisement data will be parsed and added to an partector_ble_device object.
        """
        for device, values in self._devices_to_check.items():
            if values[0] not in self._partector_clients:
                self._partector_clients[values[0]] = PartectorBleDevice(values[0])

            client = BleakClient(device, self._disconnected_callback, timeout=1)

            try:
                await client.connect()
            except asyncio.TimeoutError:
                self._add_old_device_data(values)
                continue

            my_ble_device = self._partector_clients[values[0]]
            if my_ble_device.ble_client:
                return None

            await client.start_notify(self.CHAR_STD, my_ble_device.callback_std)
            await client.start_notify(self.CHAR_AUX, my_ble_device.callback_aux)
            await client.start_notify(self.CHAR_READ, my_ble_device.callback_read)
            await client.start_notify(self.CHAR_SIZE_DIST, my_ble_device.callback_size_dist)

            my_ble_device.ble_client = client

            logger.info(
                f"Connected to {device.name} {device.address} with serial number {values[0]}"
            )

        self._devices_to_check.clear()

    def _add_old_device_data(self, values: tuple[int, AdvertisementData]) -> None:
        serial_number = values[0]
        adv = values[1]

        self._partector_clients[serial_number]._add_old_format_data(adv, serial_number)

    async def _disconnect(self) -> None:
        """Disconnect from all connected devices.

        This is done with range in for loop because client disconnect removes the client from the dict.
        A while loop would not be safe because if something goes wrong the thread would never close.
        """
        devices = list(self._partector_clients)
        for device in devices:
            logger.debug(f"Disconnecting from {device}")
            if self._partector_clients[device].ble_client is not None:
                await self._partector_clients[device].ble_client.disconnect()  # type: ignore

        logger.info("Disconnected from all BLE devices")

    async def _detection_callback(self, device: BLEDevice, data: AdvertisementData) -> None:
        """Handles all the callbacks from the BleakScanner used in the scan method.

        Args:
            device (BLEDevice): Bleak BLEDevice object
            data (AdvertisementData): Bleak AdvertisementData object
        """
        if device.name not in ["P2", "PartectorBT"]:
            return None

        _, sn = PartectorBleDevice.get_naneos_adv(data)

        if sn is None:  # return None if no serial number is found
            return None

        # return None if serial number is not in the list of serial numbers
        if self._serial_numbers is not None and sn not in self._serial_numbers:
            return None

        connected = [x.serial_number for x in self._partector_clients.values()]
        if sn in connected:
            return None

        logger.info(f"Found {device.name} {device.address} with serial number {sn}")
        self._devices_to_check[device] = (sn, data)

    def _disconnected_callback(self, client: BleakClient) -> None:
        """Removes the client from the _connected_clients dict when it disconnects.

        Args:
            client (BleakClient): BleakClient object
        """
        serial_number = next(
            x.serial_number for x in self._partector_clients.values() if x.ble_client == client
        )
        logger.debug(f"Disconnected from {serial_number}")
        self._partector_clients.pop(serial_number)


if __name__ == "__main__":
    partector_ble = PartectorBle(serial_numbers=[8112, 8446])
    partector_ble.start()
    time.sleep(5)
    partector_ble.send_command(8112, "N?")
    time.sleep(15)
    # print(partector_ble._partector_clients[8112]._data_queue)
    partector_ble.event.set()
    partector_ble.join()
